%%
%   Demo code for the synchronization method described in:
%  
%   "Multiview Registration via Graph Diffusion of Dual Quaternions"
%   A.Torsello, E.Rodola, and A.Albarelli
%   Proc. CVPR 2011
%  
%   To compile the mex code, simply run "mex diffuse_dualquat.cpp"
%  
%   Emanuele Rodola <emanuele.rodola@usi.ch>
%   USI Lugano, July 2016

clear all
close all
clc

addpath('./mex/')

% -------------------------------------------------------------------------
% Load full model and range maps
% -------------------------------------------------------------------------

full_model = load_off('./data/model.off', true);

n_rangemaps = 9;

rangemaps = cell(1, n_rangemaps);
abs_motions_gt = cell(1, n_rangemaps); % ground-truth absolute motions

for i=1:n_rangemaps
    rangemaps{i} = load_off(sprintf('./data/view%d.off', i), true);
    tmp = load(sprintf('./data/view%d.gt.txt', i));
    abs_motions_gt{i} = tmp(1:end-1,:);
end

% -------------------------------------------------------------------------
% Load input pairwise motions.
%
% This is an example with ring topology. The pairwise motions were computed
% as a perturbation of ground-truth, so as to simulate the output of some
% ICP-like algorithm.
% -------------------------------------------------------------------------

pairs = [0:8 ; mod(1:9,9)]; % NOTE: these must be 0-based, in *double* format
pairwise_motions = [];

for i=1:size(pairs,2)
    RT = load(sprintf('./data/views%d-%d.txt', pairs(1,i), pairs(2,i)));
    pairwise_motions(i).i1 = double( pairs(1,i) );
    pairwise_motions(i).i2 = double( pairs(2,i) );
    pairwise_motions(i).R = double( RT(1:3,1:3) );
    pairwise_motions(i).T = double( RT(end,:) );
end

% -------------------------------------------------------------------------
% Run dual quaternion diffusion.
%
% NOTE: The output is a set of *absolute* rigid motions, even though the
% input is pairwise.
% -------------------------------------------------------------------------

% clc
% cd('./mex/')
% mex -v ./diffuse_dualquat.cpp
% cd('..')

absolute_diffused = diffuse_dualquat(pairwise_motions, n_rangemaps);

figure

subplot(131), hold on
for i=1:n_rangemaps
    plot_scalar_map(rangemaps{i}, i*ones(rangemaps{i}.n,1)), shading interp; axis off
end
title('initial')

subplot(133)
plot_scalar_map(full_model, ones(full_model.n,1)), shading interp; axis off
light; camlight head; lighting phong
title('ground truth model')

subplot(132), hold on
for i=1:n_rangemaps
    M = rangemaps{i};
    M.VERT = M.VERT * absolute_diffused(i).R' + repmat(absolute_diffused(i).T, M.n, 1);
    plot_scalar_map(M, i*ones(M.n,1)), shading interp; axis off
end
title('synchronized')
